//
//	texture.cpp - source file
//
//	David Henry - tfc_duke@club-internet.fr
//


#pragma warning( disable : 4005 ) // APIENTRY and WINGDIAPI macro redefinition

#include	<GL/glfw.h>
#include	<GL/glext.h>
#include	<iostream>
#include	<math.h>

#include	"texture.h"
#include	"bmp.h"



/////////////////////////////////////////////////
//				CTextureManager				   //
/////////////////////////////////////////////////


// ----------------------------------------------
// Initialize() - cre une texture d'chiquier
// pour la texture par dfaut.
// ----------------------------------------------

void CTextureManager::Initialize( void )
{
	// vide la base de donnes
	ReleaseTextures();

	// c'est la premire texture charge en mmoire. Si
	// une texture ne peut tre lue ou charge, on utilise
	// celle si  la place.

	std::cout << "CTextureManager : Texture list is empty, creating default texture..." << std::endl;
	GLuint id;

	// cration et initialisation de la texture
	glGenTextures( 1, &id );
	glBindTexture( GL_TEXTURE_2D, id );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );

	// ajout  la base de donne
	m_kDataBase[ "default" ] = id;

	std::cout << "CTextureManager : creating default texture, id [" << id << "]" << std::endl;


	// cration d'une texture chiquier pour la texture par dfaut
	int i, j, c;			// variables de contrle
	GLubyte *checker;		// donnes texels

	checker = new GLubyte[ 64 * 64 * 4 ];

	// construction de l'chiquier
	for( i = 0; i < 64; i++ )
	{
		for( j = 0; j < 64; j++ )
		{
			c = ( !(i & 8) ^ !(j & 8)) * 255;

			checker[ (i * 256) + (j * 4) + 0 ] = (GLubyte)c;
			checker[ (i * 256) + (j * 4) + 1 ] = (GLubyte)c;
			checker[ (i * 256) + (j * 4) + 2 ] = (GLubyte)c;
			checker[ (i * 256) + (j * 4) + 3 ] = (GLubyte)255;
		}
	}

	glTexImage2D( GL_TEXTURE_2D, 0, 4, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, checker );
	delete [] checker;
}



// ----------------------------------------------
// LoadTexture() - charge une texture. Vrifie
// si elle n'est pas dj stocke en mmoire. Si
// c'est le cas, la fonction retourne l'id de la
// texture dj stocke. Sinon, elle retourne
// l'id de la nouvelle texture.
// ----------------------------------------------

GLuint CTextureManager::LoadTexture( std::string szFilename )
{
	GLuint		id = 0;			// id de la texture  retourner
	GLubyte		*pixels = 0;	// tableau de pixels de la texture
	int			width, height;	// dimensions de la texture


	// on tente d'obtenir la texture depuis la base de donnes.
	// Si elle a dj t charge au pralable, on retourne son
	// id, sinon l'id retourne est celle de la texture par dfaut.
	// On tente alors de la charger dans la base.
	id = GetTexture( szFilename );

	if( id != GetTexture( "default" ) )
	{
		// la texture a dj t charge
		std::cout << "CTextureManager : " << szFilename << " is already stored, id [" << id << "]" << std::endl;
	}
	else
	{
		// on tente de charger la texture
		if( LoadFileBMP( szFilename.c_str(), &pixels, &width, &height, false ) > 0 )
		{
			// cration et initialisation de la texture opengl
			glGenTextures( 1, &id );
			glBindTexture( GL_TEXTURE_2D, id );

			glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
			glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
			glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
			glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

			glTexImage2D( GL_TEXTURE_2D, 0, 4, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, pixels );

			// cration d'une nouvelle texture et ajout  la base de donnes
			m_kDataBase[ szFilename ] = id;

			std::cout << "CTextureManager : " << szFilename << " successfully loaded!" << std::endl;
			std::cout << " # creating new texture, id [" << id << "], (" << width<< " * " << height<< ")" << std::endl;
		}
		else
		{
			// impossible de charger la texture, texture par dfault utilise
			std::cout << "CTextureManager : failed to load " << szFilename << "! giving default texture id [" << m_kDataBase[ "default" ] << "]" << std::endl;
		}

		// dsallocation de mmoire - OpenGL possde sa propre copie de l'image
		if( pixels )
			delete [] pixels;
	}

	return id;
}



// ----------------------------------------------
// GetTexture() - retourne l'id de la texture
// filename si dj charge.
// ----------------------------------------------

GLuint CTextureManager::GetTexture( std::string szFilename )
{
	if( (m_kItor = m_kDataBase.find( szFilename )) != m_kDataBase.end() )
	{
		return (*m_kItor).second;
	}
	else
	{
		return m_kDataBase[ "default" ];
	}
}



// ----------------------------------------------
// GetTextureName() - retourne un pointeur sur le
// nom de l'image d'id spcifie.
// ----------------------------------------------

char *CTextureManager::GetTextureName( GLuint id )
{
	if( glIsTexture( id ) )
	{
		for( m_kItor = m_kDataBase.begin(); m_kItor != m_kDataBase.end(); ++m_kItor )
		{
			if( (*m_kItor).second == id )
			{
				return (char *)((*m_kItor).first.c_str());
			}
		}
	}

	// la texture n'a pas t trouve
	return (char *)"failed";
}



// ----------------------------------------------
// ReleaseTextures() - dtruit toutes les textures
// stockes dans la base de donnes.
// ----------------------------------------------

void CTextureManager::ReleaseTextures( void )
{
	std::cout << "CTextureManager : cleaning all textures!!! size = " << m_kDataBase.size();

	// on dtruit chaque texture de la base de donnes du manager
	for( m_kItor = m_kDataBase.begin(); m_kItor != m_kDataBase.end(); )
	{
		glDeleteTextures( 1, &(*m_kItor).second );
		m_kDataBase.erase( m_kItor++ );
	}

	std::cout << " // cleaned! size = " << m_kDataBase.size() << std::endl;
}



// ----------------------------------------------
// DeleteTexture() - dtruit la texture szName.
// ----------------------------------------------

void CTextureManager::DeleteTexture( std::string szName )
{
	// on recherche l'id de la texture  dtruire
	if( (m_kItor = m_kDataBase.find( szName )) != m_kDataBase.end() )
	{
		std::cout << "CTextureManager : erasing " << szName << std::endl;

		// destruction de la texture et retrait de l'id
		// de la base de donnes du manager
		glDeleteTextures( 1, &(*m_kItor).second );
		m_kDataBase.erase( m_kItor );
	}
	else
	{
		std::cout << "CTextureManager : couldn't erase " << szName << std::endl;
	}
}



// ----------------------------------------------
// DeleteTexture() - dtruit la texture de l'id
// spcifie.
// ----------------------------------------------

void CTextureManager::DeleteTexture( GLuint id )
{
	if( glIsTexture( id ) )
	{
		for( m_kItor = m_kDataBase.begin(); m_kItor != m_kDataBase.end(); ++m_kItor )
		{
			if( (*m_kItor).second == id )
			{
				std::cout << "CTextureManager : erasing " << (*m_kItor).first<< " [" << (*m_kItor).second << "], size = " << m_kDataBase.size() << std::endl;

				// destruction de la texture et retrait de l'id
				// de la base de donnes du manager
				glDeleteTextures( 1, &(*m_kItor).second );
				m_kDataBase.erase( m_kItor );

				return;
			}
		}
	}
}
